# SPDX-License-Identifier: MIT

from enum import IntEnum
from m1n1.utils import *


class R_USB2PHY_USBCTL(Register32):
    MODE_HOST = 1
    MODE_ISOLATION = 2


class R_USB2PHY_CTL(Register32):
    RESET = 0
    PORT_RESET = 1
    APB_RESETN = 2
    SIDDQ = 3


class R_USB2PHY_SIG(Register32):
    VBUSDET_FORCE_VAL = 0
    VBUSDET_FORCE_EN = 1
    VBUSVLDEXT_FORCE_VAL = 2
    VBUSVLDEXT_FORCE_EN = 3
    MODE_HOST = 19, 12


class R_USB2PHY_MISCTUNE(Register32):
    APBCLK_GATE_OFF = 29
    REFCLK_GATE_OFF = 30


class Usb2PhyRegs(RegMap):
    USB2PHY_USBCTL = 0x00, R_USB2PHY_USBCTL
    USB2PHY_CTL = 0x04, R_USB2PHY_CTL
    USB2PHY_SIG = 0x08, R_USB2PHY_SIG
    USB2PHY_MISCTUNE = 0x1C, R_USB2PHY_MISCTUNE


class R_AUSPLL_DCO_EFUSE_SPARE(Register32):
    RODCO_ENCAP_EFUSE = 10, 9
    RODCO_BIAS_ADJUST_EFUSE = 14, 12


class R_AUSPLL_FRACN_CAN(Register32):
    DLL_START_CAPCODE = 18, 17


class R_AUSPLL_CLKOUT_DTC_VREG(Register32):
    DTC_VREG_ADJUST = 16, 14


class R_AUS_COMMON_SHIM_BLK_VREG(Register32):
    VREG_TRIM = 6, 2


class R_CIO3PLL_CLK_CTRL(Register32):
    PCLK_EN = 1
    REFCLK_EN = 5


class R_CIO3PLL_DCO_NCTRL(Register32):
    DCO_COARSEBIN_EFUSE0 = 6, 0
    DCO_COARSEBIN_EFUSE1 = 23, 17


class R_CIO3PLL_FRACN_CAN(Register32):
    DLL_CAL_START_CAPCODE = 18, 17


class R_CIO3PLL_DTC_VREG(Register32):
    DTC_VREG_ADJUST = 16, 14


class E_ACIOPHY_CROSSBAR_PROTOCOL(IntEnum):
    USB4 = 0
    USB3 = 5
    USB3_DP = 8
    DP = 10


class R_ACIOPHY_CROSSBAR(Register32):
    PROTOCOL_SWAPPED = 0
    PROTOCOL = 4, 1, E_ACIOPHY_CROSSBAR_PROTOCOL
    DPMODE = 17, 15


class E_ACIOPHY_LANE_MODE(IntEnum):
    USB4 = 0
    USB3 = 1
    DP = 2
    OFF = 3


class R_ACIOPHY_LANE_MODE(Register32):
    RX0 = 2, 0, E_ACIOPHY_LANE_MODE
    TX0 = 5, 3, E_ACIOPHY_LANE_MODE
    RX1 = 8, 6, E_ACIOPHY_LANE_MODE
    TX1 = 11, 9, E_ACIOPHY_LANE_MODE


class R_ATCPHY_POWER(Register32):
    SLEEP_SMALL = 0
    SLEEP_BIG = 1
    CLAMP_EN = 2
    APB_RESET_N = 3
    PHY_RESET_N = 4


class R_ATCPHY_MISC(Register32):
    RESET_N = 0
    LANE_SWAP = 2


class AtcPhyRegs(RegMap):
    ACIOPHY_CFG0 = 0x08, Register32
    ACIOPHY_LANE_MODE = 0x48, R_ACIOPHY_LANE_MODE
    ACIOPHY_CROSSBAR = 0x4C, R_ACIOPHY_CROSSBAR
    ACIOPHY_BIST_EN = 0x84, Register32
    ACIOPHY_BIST_OV = 0x8C, Register32
    ACIOPHY_BIST_CFG0 = 0x90, Register32
    ACIOPHY_BIST_STAT = 0x9C, Register32
    ACIOPHY_BIST_RESET = 0xA8, Register32
    ACIOPHY_BIST_CFG1 = 0xAC, Register32
    ACIOPHY_SLEEP_CTRL = 0x1B0, Register32

    AUS_COMMON_SHIM_BLK_VREG = 0x0A04, R_AUS_COMMON_SHIM_BLK_VREG

    AUSPLL_FSM_CTRL = 0x1014, Register32
    AUSPLL_CMD_OVERRIDE = 0x2000, Register32
    AUSPLL_CLKOUT_DTC_VREG = 0x2220, R_AUSPLL_CLKOUT_DTC_VREG
    AUSPLL_DCO_EFUSE_SPARE = 0x222C, R_AUSPLL_DCO_EFUSE_SPARE
    AUSPLL_FRACN_CAN = 0x22A4, R_AUSPLL_FRACN_CAN

    CIO3PLL_CLK_CTRL = 0x2A00, R_CIO3PLL_CLK_CTRL
    CIO3PLL_DTC_VREG = 0x2A20, R_CIO3PLL_DTC_VREG
    CIO3PLL_DCO_NCTRL = 0x2A38, R_CIO3PLL_DCO_NCTRL
    CIO3PLL_FRACN_CAN = 0x2AA4, R_CIO3PLL_FRACN_CAN

    ATCPHY_POWER_CTRL = 0x20000, R_ATCPHY_POWER
    ATCPHY_POWER_STAT = 0x20004, R_ATCPHY_POWER
    ATCPHY_MISC = 0x20008, R_ATCPHY_MISC

    # tunable MMIO ranges, accessed based on ADT tunable properties
    LN0_AUSPMA_RX_TOP = irange(0x9000, 0x1000 // 4, 4), Register32
    LN0_AUSPMA_RX_EQ = irange(0xA000, 0x1000 // 4, 4), Register32
    LN0_AUSPMA_RX_SHM = irange(0xB000, 0x1000 // 4, 4), Register32
    LN0_AUSPMA_TX_TOP = irange(0xC000, 0x1000 // 4, 4), Register32
    LN1_AUSPMA_RX_TOP = irange(0x10000, 0x1000 // 4, 4), Register32
    LN1_AUSPMA_RX_EQ = irange(0x11000, 0x1000 // 4, 4), Register32
    LN1_AUSPMA_RX_SHM = irange(0x12000, 0x1000 // 4, 4), Register32
    LN1_AUSPMA_TX_TOP = irange(0x13000, 0x1000 // 4, 4), Register32

    ACIOPHY_TOP_TUNABLE_118 = 0x118, Register32
    ACIOPHY_TOP_TUNABLE_11c = 0x11C, Register32
    ACIOPHY_TOP_TUNABLE_124 = 0x124, Register32
    ACIOPHY_TOP_TUNABLE_a00 = 0xA00, Register32
    ACIOPHY_TOP_TUNABLE_808 = 0x808, Register32

    AUSPLL_TOP_TUNABLE_2088 = 0x2088, Register32
    AUSPLL_TOP_TUNABLE_2094 = 0x2094, Register32
    AUSPLL_TOP_TUNABLE_20a0 = 0x20A0, Register32
    AUSPLL_TOP_TUNABLE_20ac = 0x20AC, Register32
    AUSPLL_TOP_TUNABLE_20b8 = 0x20B8, Register32

    CIO3PLL_TOP_TUNABLE_10 = 0x2810, Register32
    CIO3PLL_TOP_TUNABLE_88 = 0x2888, Register32
    CIO3PLL_TOP_TUNABLE_94 = 0x2894, Register32

    CIO3PLL_CORE_TUNABLE_1c = 0x2A1C, Register32
    CIO3PLL_CORE_TUNABLE_28 = 0x2A28, Register32
    CIO3PLL_CORE_TUNABLE_9c = 0x2A9C, Register32

    AUSPLL_CORE_TUNABLE_78 = 0x2278, Register32
